from distutils.util import strtobool
import os
import warnings

from .collector import get_collector_path, get_hostname


def set_config(**config):
    """
    Returns IOpipe configuration options, setting defaults as necessary.
    """
    config.setdefault('debug', bool(strtobool(os.getenv('IOPIPE_DEBUG', 'false'))))
    config.setdefault('enabled', bool(strtobool(os.getenv('IOPIPE_ENABLED', 'true'))))
    config.setdefault('host', get_hostname())
    config.setdefault('install_method', os.getenv('IOPIPE_INSTALL_METHOD', 'manual'))
    config.setdefault('network_timeout', os.getenv('IOPIPE_NETWORK_TIMEOUT', 5000))
    config.setdefault('path', get_collector_path())
    config.setdefault('plugins', [])
    config.setdefault('sync_http', False)
    config.setdefault('timeout_window', os.getenv('IOPIPE_TIMEOUT_WINDOW', 500))
    config.setdefault('token', os.getenv('IOPIPE_TOKEN') or os.getenv('IOPIPE_CLIENTID') or '')

    if 'client_id' in config:
        config['token'] = config.pop('client_id')

    if 'url' in config:
        url = config.pop('url')

        config['host'] = get_hostname(url)
        config['path'] = get_collector_path(url)

    if '.' in str(config['network_timeout']):
        warnings.warn("IOpipe's 'network_timeout' is now in milliseconds, expressed as an integer")

    try:
        config['debug'] = bool(config['debug'])
    except ValueError:
        config['debug'] = False

    try:
        config['network_timeout'] = int(config['network_timeout']) / 1000.0
    except ValueError:
        config['network_timeout'] = 5

    if '.' in str(config['timeout_window']):
        warnings.warn("IOpipe's 'timeout_window' is now in milliseconds, expressed as an integer")

    try:
        config['timeout_window'] = int(config['timeout_window']) / 1000.0
    except ValueError:
        config['timeout_window'] = 0.5

    return config



from iopipe.plugins import Plugin

from .marker import Marker
from .timeline import Timeline
from .util import add_timeline_measures


class TracePlugin(Plugin):
    name = 'trace'
    version = '1.1.0'
    homepage = 'https://github.com/iopipe/iopipe-python#trace-plugin'
    enabled = True

    def __init__(self, auto_measure=True):
        self.auto_measure = auto_measure
        self.timeline = Timeline()

    def pre_setup(self, iopipe):
        pass

    def post_setup(self, iopipe):
        pass

    def pre_invoke(self, event, context):
        context.iopipe.register('mark', Marker(self.timeline, context))

    def post_invoke(self, event, context):
        context.iopipe.unregister('mark')

    def pre_report(self, report):
        if self.auto_measure:
            add_timeline_measures(self.timeline)
        report.report['performanceEntries'] = [e._asdict() for e in self.timeline.get_entries()]

    def post_report(self, report):
        self.timeline = Timeline()



import json
import logging
import os
import platform
import sys
import time
import traceback

from . import constants
from .monotonic import monotonic
from .plugins import get_plugin_meta
from .send_report import send_report

if sys.platform.startswith('linux'):
    from . import system
else:
    from . import mock_system as system

logger = logging.getLogger(__name__)


class Report(object):
    """
    The report of system status
    """

    def __init__(self, client, context):
        """
        Instantiates a new IOpipe report.

        :param client: The IOpipe client.
        :param context: The AWS Lambda context.
        """
        self.start_time = monotonic()
        self.sent = False
        self.stat_start = system.read_pid_stat('self')

        self.client = client
        self.config = client.config
        self.context = context

        self.custom_metrics = []
        self.labels = set()
        self.plugins = get_plugin_meta(self.config['plugins'])

        self.report = {
            'client_id': self.config['token'],
            'coldstart': constants.COLDSTART,
            'custom_metrics': self.custom_metrics,
            'environment': {
                'agent': {
                    'load_time': constants.MODULE_LOAD_TIME,
                    'runtime': 'python',
                    'version': constants.VERSION,
                },
                'runtime': {
                    'name': platform.python_implementation(),
                    'version': platform.python_version()
                },
                # DEPRECATED: the following key will be removed in favor of
                # the 'runtime' in the future
                'python': {
                    'version': platform.python_version(),
                },
                'host': {},
                'os': {
                    'linux': {},
                },
            },
            'errors': {},
            'installMethod': self.config.get('install_method'),
            'plugins': self.plugins,
            'processId': constants.PROCESS_ID,
            'timestamp': int(time.time() * 1000),
        }

        if constants.COLDSTART is True:
            constants.COLDSTART = False
            self.labels.add('@iopipe/coldstart')

    def extract_context_data(self):
        """
        Returns the contents of a AWS Lambda context.

        :returns: A dict of relevant context data.
        :rtype: dict
        """
        data = {}
        for k, v in {
                # camel case names in the report to align with AWS standards
                'functionName': 'function_name',
                'functionVersion': 'function_version',
                'memoryLimitInMB': 'memory_limit_in_mb',
                'invokedFunctionArn': 'invoked_function_arn',
                'awsRequestId': 'aws_request_id',
                'logGroupName': 'log_group_name',
                'logStreamName': 'log_stream_name',
        }.items():
            if hasattr(self.context, v):
                data[k] = getattr(self.context, v)
        if hasattr(self.context, 'invoked_function_arn') and \
                'AWS_SAM_LOCAL' in os.environ:
            data['invokedFunctionArn'] = \
                'arn:aws:lambda:local:0:function:%s' % \
                data.get('functionName', 'unknown')
        if hasattr(self.context, 'get_remaining_time_in_millis') and \
                callable(self.context.get_remaining_time_in_millis):
            data['getRemainingTimeInMillis'] = self.context.get_remaining_time_in_millis()
        data['traceId'] = os.getenv('_X_AMZN_TRACE_ID', '')
        return data

    def retain_error(self, error, frame=None):
        """
        Adds details of an error to the report.

        :param error: The error exception to add to the report.
        """
        if frame is None:
            stack = traceback.format_exc()
            self.labels.add('@iopipe/error')
        else:
            stack = '\n'.join(traceback.format_stack(frame))
            self.labels.add('@iopipe/timeout')
        details = {
            'name': type(error).__name__,
            'message': '{}'.format(error),
            'stack': stack,
        }
        self.report['errors'] = details

    def prepare(self, error=None, frame=None):
        """
        Prepare the report to be sent to IOpipe.

        :param error: An optional error to add to report.
        :param frame: An optional stack frame to add to report in the event of a timeout.
        """
        if error:
            self.retain_error(error, frame)

        self.report['environment']['host']['boot_id'] = system.read_bootid()

        # convert labels to list for sending
        self.report['labels'] = list(self.labels)

        meminfo = system.read_meminfo()

        self.report.update({
            'aws': self.extract_context_data(),
            'timestampEnd': int(time.time() * 1000),
        })

        self.report['environment']['os'].update({
            'cpus': system.read_stat(),
            'freemem': meminfo['MemFree'],
            'hostname': system.read_hostname(),
            'totalmem': meminfo['MemTotal'],
            'usedmem': meminfo['MemTotal'] - meminfo['MemFree'],
        })

        self.report['environment']['os']['linux']['pid'] = {
            'self': {
                'stat': system.read_pid_stat('self'),
                'stat_start': self.stat_start,
                'status': system.read_pid_status('self'),
            },
        }

        self.report['disk'] = system.read_disk()

        self.report['duration'] = int((monotonic() - self.start_time) * 1e9)

    def send(self):
        """
        Sends the report to IOpipe.
        """
        if self.sent is True:
            return
        self.sent = True

        logger.debug('Sending report to IOpipe:')
        logger.debug(json.dumps(self.report, indent=2, sort_keys=True))

        self.client.submit_future(send_report, self.report, self.config)



import logging
import os
import time

try:
    import requests
except ImportError:
    from botocore.vendored import requests

from .collector import SUPPORTED_REGIONS

logger = logging.getLogger(__name__)


def get_signer_hostname():
    """
    Returns the IOpipe signer hostname for a region

    :returns: The signer hostname
    :rtype str
    """
    region = os.getenv('AWS_REGION', '')
    region = region if region and region in SUPPORTED_REGIONS else 'us-west-2'
    return 'signer.{region}.iopipe.com'.format(region=region)


def get_signed_request(token, context, extension):
    """
    Returns a signed request URL from IOpipe

    :param token: The IOpipe token.
    :param context: The AWS context to request a signed URL
    :param extension: The extension of the file to sign.
    :returns: A signed request URL
    :rtype: str
    """
    url = 'https://{hostname}/'.format(hostname=get_signer_hostname())

    try:
        logger.debug('Requesting signed request URL from %s', url)
        response = requests.post(
            url,
            json={
                'arn': context.invoked_function_arn,
                'requestId': context.aws_request_id,
                'timestamp': int(time.time() * 1000),
                'extension': extension
            },
            headers={
                'Authorization': token,
            })
        response.raise_for_status()
    except Exception as e:
        logger.debug('Error requesting signed request URL: %s', e)
        if hasattr(e, 'response'):
            logger.debug(e.response.content)
    else:
        response = response.json()
        logger.debug('Signed request URL received for %s', response['url'])
        return response



from .util import collect_all_keys, get_value, has_key, slugify


class EventType(object):
    keys = []
    exclude_keys = []
    required_keys = []
    source = None

    def __init__(self, event):
        self.event = event

    @property
    def slug(self):
        return slugify(self.type)

    def has_required_keys(self):
        return all(has_key(self.event, key) for key in self.required_keys)

    def collect(self):
        if self.keys == 'all':
            return collect_all_keys(self.event, '@iopipe/event-info.%s' % self.type, self.exclude_keys)
        event_info = {}
        for key in self.keys:
            if isinstance(key, tuple):
                old_key, new_key = key
            else:
                old_key = new_key = key
            value = get_value(self.event, old_key)
            if value is not None:
                event_info['@iopipe/event-info.%s.%s' % (self.type, new_key)] = value
        event_info['@iopipe/event-info.eventType'] = self.type
        if self.source:
            event_info['@iopipe/event-info.eventType.source'] = self.source
        return event_info


class AlexaSkill(EventType):
    type = 'alexaSkill'
    keys = 'all'
    exclude_keys = [
        'context.System.apiAccessToken',
        'context.System.user.accessToken',
        'context.System.user.permissions.consentToken',
    ]
    required_keys = [
        'context.System',
        'request.requestId',
        'session.attributes',
        'session.user',
    ]


class ApiGateway(EventType):
    type = 'apiGateway'
    keys = [
        'httpMethod',
        'path',
        'requestContext.accountId',
        'requestContext.httpMethod',
        'requestContext.identity.userAgent',
        'requestContext.requestId',
        'requestContext.resourcePath',
        'requestContext.stage',
        'resource',
    ]
    required_keys = [
        'headers',
        'httpMethod',
        'path',
        'requestContext',
        'resource',
    ]


class CloudFront(EventType):
    type = 'cloudFront'
    keys = [
        'Records[0].cf.config.distributionId',
        'Records[0].cf.request.clientIp',
        'Records[0].cf.request.headers.host[0].value',
        'Records[0].cf.request.headers.["user-agent"][0].value',
        'Records[0].cf.request.method',
        'Records[0].cf.request.uri',
    ]
    required_keys = ['Records[0].cf']


class Firehose(EventType):
    type = 'firehose'
    keys = [
        'deliveryStreamArn',
        'region',
    ]
    required_keys = [
        'deliveryStreamArn',
        'records[0]',
        'records[0].kinesisRecordMetadata',
    ]


class Kinesis(EventType):
    type = 'kinesis'
    keys = [
        'Records.length',
        'Records[0].awsRegion',
        'Records[0].eventSourceARN',
    ]
    required_keys = [
        'Records[0].eventVersion',
        'Records[0].eventSource',
    ]

    def has_required_keys(self):
        return super(Kinesis, self).has_required_keys() and \
            get_value(self.event, 'Records[0].eventVersion') == '1.0' and \
            get_value(self.event, 'Records[0].eventSource') == 'aws:kinesis'


class S3(EventType):
    type = 's3'
    keys = [
        'Records[0].awsRegion',
        'Records[0].eventName',
        'Records[0].eventTime',
        'Records[0].requestParameters.sourceIPAddress',
        'Records[0].responseElements["x-amz-id-2"]',
        'Records[0].responseElements["x-amz-request-id"]',
        'Records[0].s3.bucket.arn',
        'Records[0].s3.bucket.name',
        'Records[0].s3.object.key',
        'Records[0].s3.object.sequencer',
        'Records[0].s3.object.size',
        'Records[0].userIdentity.principalId',
    ]
    required_keys = [
        'Records[0].eventVersion',
        'Records[0].eventSource',
    ]

    def has_required_keys(self):
        return super(S3, self).has_required_keys() and \
            get_value(self.event, 'Records[0].eventVersion') == '2.0' and \
            get_value(self.event, 'Records[0].eventSource') == 'aws:s3'


class Scheduled(EventType):
    type = 'scheduled'
    keys = [
        'account',
        'id',
        'region',
        'resources[0]',
        'time',
    ]
    required_keys = ['source']

    def has_required_keys(self):
        return super(Scheduled, self).has_required_keys() and get_value(self.event, 'source') == 'aws.events'


class ServerlessLambda(EventType):
    type = 'apiGateway'
    source = 'slsIntegrationLambda'
    keys = [
        ('headers.["X-Amz-Cf-Id"]', 'headers.X-Amz-Cf-Id'),
        ('headers.["X-Amzn-Trace-Id"]', 'headers.X-Amzn-Trace-Id'),
        ('identity.accountId', 'requestContext.accountId'),
        ('identity.userAgent', 'requestContext.identity.userAgent'),
        ('method', 'httpMethod'),
        ('method', 'requestContext.httpMethod'),
        ('stage', 'requestContext.stage'),
    ]
    required_keys = [
        'identity.userAgent',
        'identity.sourceIp',
        'identity.accountId',
    ]


class SNS(EventType):
    type = 'sns'
    keys = [
        'Records[0].EventSubscriptionArn',
        'Records[0].Sns.MessageId',
        'Records[0].Sns.Signature',
        'Records[0].Sns.SignatureVersion',
        'Records[0].Sns.SigningCertUrl',
        'Records[0].Sns.UnsubscribeUrl',
        'Records[0].Sns.Subject',
        'Records[0].Sns.Timestamp',
        'Records[0].Sns.TopicArn',
        'Records[0].Sns.Type',
    ]
    required_keys = [
        'Records[0].eventVersion',
        'Records[0].eventSource',
    ]

    def has_required_keys(self):
        return super(SNS, self).has_required_keys() and \
            get_value(self.event, 'Records[0].eventVersion') == '1.0' and \
            get_value(self.event, 'Records[0].eventSource') == 'aws:sns'


EVENT_TYPES = [AlexaSkill, ApiGateway, CloudFront, Firehose, Kinesis, S3, Scheduled, ServerlessLambda, SNS]


def metrics_for_event_type(event, context):
    for EventType in EVENT_TYPES:
        event_type = EventType(event)
        if event_type.has_required_keys():
            context.iopipe.label('@iopipe/plugin-event-info')
            context.iopipe.label('@iopipe/aws-%s' % event_type.slug)
            event_info = event_type.collect()
            [context.iopipe.metric(k, v) for k, v in event_info.items()]
            break



from __future__ import division

import random
import uuid

from .system import read_hostname  # noqa


def read_bootid():
    """
    Mocks read_bootid as this is a Linux-specific operation.
    """
    return uuid.uuid4().hex


def read_disk():
    """
    Mocks read_disk as this is a Linux-specific operation.
    """
    used = random.randint(0, 500)
    return {'totalMiB': 500, 'usedMiB': used, 'usedPercentage': round((used / 500) * 100, 2)}


def read_meminfo():
    """
    Mocks read_meminfo as this is a Linux-specific operation.
    """
    return {
        'MemTotal': random.randint(0, 999999999),
        'MemFree': random.randint(0, 999999999),
        'MemAvailable': random.randint(0, 999999999),
        'HugePages_Total': random.randint(0, 999999999),
    }


def read_pid_stat(pid):
    """
    Mocks read_pid_stat as this is a Linux-specific operation.
    """
    return {
        'utime': random.randint(0, 999999999),
        'stime': random.randint(0, 999999999),
        'cutime': random.randint(0, 999999999),
        'cstime': random.randint(0, 999999999),
    }


def read_pid_status(pid):
    """
    Mocks read_pid_status as this is a Linux-specific operation.
    """
    return {
        'FDSize': random.randint(0, 999999999),
        'Threads': random.randint(0, 99999),
        'VmRSS': random.randint(0, 999999999),
    }


def read_stat():
    """
    Mocks read_stat as this is a Linux-specific operation.
    """
    return [{
        'times': {
            'user': random.randint(0, 999999999),
            'nice': random.randint(0, 999999999),
            'sys': random.randint(0, 999999999),
            'idle': random.randint(0, 999999999),
            'irq': random.randint(0, 999999999),
        }
    }]


def read_uptime():
    """
    Mocks read_updtime as this is a Linux-specific operaiton.
    """
    return random.randint(0, 999999999)



def add_timeline_measures(timeline):
    entries = timeline.get_entries_by_type('mark')
    names = [e.name for e in entries if e.name.startswith('start:') or e.name.startswith('end:')]
    for name in names:
        if name.startswith('start:'):
            _, base_name = name.split(':', 1)
            end_name = 'end:%s' % base_name
            if end_name in names:
                timeline.measure('measure:%s' % base_name, name, end_name)



class LogWrapper(object):
    def __init__(self, logger, context):
        self.logger = logger
        self.context = context

    def __call__(self, key, value):
        self.context.iopipe.metric(key, value)

    def __getattr__(self, name):
        self.context.iopipe.label('@iopipe/plugin-logger')
        if name in ['warn', 'warning']:
            self.context.iopipe.label('@iopipe/logs-warning')
        if name in ['critical', 'error', 'exception']:
            self.context.iopipe.label('@iopipe/logs-error')
        return getattr(self.logger, name)



class Marker(object):
    def __init__(self, timeline, context):
        self.timeline = timeline
        self.context = context
        self.contexts = []

    def __call__(self, name):
        self.contexts.append(name)
        return self

    def __enter__(self):
        if self.contexts:
            self.start(self.contexts[-1])
        return self

    def __exit__(self, type, value, traceback):
        if self.contexts:
            self.end(self.contexts[-1])
            self.contexts.pop()

    def start(self, name):
        self.timeline.mark('start:%s' % name)
        self.context.iopipe.label('@iopipe/plugin-trace')

    def end(self, name):
        self.timeline.mark('end:%s' % name)

    def measure(self, name, start=None, end=None):
        self.timeline.measure('measure:%s' % name, 'start:%s' % (start or name), 'end:%s' % (end or start or name))



import os
import sys

from setuptools import find_packages, setup

try:
    from pypandoc import convert
    README = convert('README.md', 'rst')
except (ImportError, OSError):
    README = open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r').read()

install_requires = []
if sys.version_info[0] == 2:
    install_requires.append('futures')

setup(
    name='iopipe',
    version='1.6.1',
    description='IOpipe agent for serverless Application Performance Monitoring',
    long_description=README,
    author='IOpipe',
    author_email='dev@iopipe.com',
    url='https://github.com/iopipe/iopipe-python',
    packages=find_packages(exclude=('tests', 'tests.*',)),
    extras_require={
        'dev': ['jmespath>=0.7.1,<1.0.0', 'requests'],
    },
    install_requires=install_requires,
    setup_requires=['pytest-runner'],
    tests_require=['flake8', 'jmespath>=0.7.1,<1.0.0', 'mock', 'pytest', 'pytest-benchmark', 'requests'],
    zip_safe=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ])



from __future__ import division

import os
import socket

MB_FACTOR = float(1 << 20)


def read_bootid():
    """
    Returns the system boot id.

    :returns: The system bood it.
    :rtype: str
    """
    with open('/proc/sys/kernel/random/boot_id') as bootid_file:
        return bootid_file.readline()


def read_disk():
    """
    Returns disk usage for /tmp

    :returns: Disk usage (total, used, percentage used)
    :rtype dict
    """
    s = os.statvfs('/tmp')
    return {
        # This should report as 500MB, if not may need to be hardcoded
        # https://aws.amazon.com/lambda/faqs/
        'totalMiB': (s.f_blocks * s.f_bsize) / MB_FACTOR,
        'usedMiB': ((s.f_blocks - s.f_bfree) * s.f_frsize) / MB_FACTOR,
        'usedPercentage': round((((s.f_blocks - s.f_bfree) * s.f_frsize) / (s.f_blocks * s.f_bsize)) * 100, 2)
    }


def read_hostname():
    """
    Returns the system hostname.

    :returns: The system hostname.
    :rtype: str
    """
    return socket.gethostname()


def read_meminfo():
    """
    Returns system memory usage information.

    :returns: The system memory usage.
    :rtype: dict
    """
    data = {}
    with open('/proc/meminfo') as meminfo_file:
        for row in meminfo_file:
            line = row.split(":")
            # Example content:
            # MemTotal:                3801016 kB
            # MemFree:                 1840972 kB
            # MemAvailable:        3287752 kB
            # HugePages_Total:             0
            data[line[0]] = int(line[1].lstrip().rstrip(" kB\n")) * 1024
    return data


def read_pid_stat(pid='self'):
    """
    Returns system process stat information.

    :param pid: The process ID.
    :returns: The system stat information.
    :rtype: dict
    """
    with open('/proc/%s/stat' % (pid, )) as f:
        stat = f.readline().split(' ')
    return {
        'utime': int(stat[13]),
        'stime': int(stat[14]),
        'cutime': int(stat[15]),
        'cstime': int(stat[16]),
    }


def read_pid_status(pid='self'):
    """
    Returns the system process sstatus.

    :param pid: The process ID.
    :returns: The system process status.
    :rtype: dict
    """
    data = {}
    with open("/proc/%s/status" % (pid, )) as status_file:
        for row in status_file:
            line = row.split(':')
            status_value = line[1].rstrip('\t\n kB').lstrip()
            if line[0] in ['FDSize', 'Threads', 'VmRSS']:
                try:
                    data[line[0]] = int(status_value)
                except ValueError:
                    data[line[0]] = status_value
    return data


def read_stat():
    """
    Returns the system stat information.

    :returns: The system stat information.
    :rtype: list
    """
    data = []
    with open('/proc/stat') as stat_file:
        for line in stat_file:
            cpu_stat = line.split(' ')
            if cpu_stat[0][:3] != "cpu":
                break
            # First cpu line is aggregation of following lines, skip it
            if len(cpu_stat[0]) == 3:
                continue
            data.append({
                'times': {
                    'user': int(cpu_stat[1]),
                    'nice': int(cpu_stat[2]),
                    'sys': int(cpu_stat[3]),
                    'idle': int(cpu_stat[4]),
                    'irq': int(cpu_stat[6])
                }
            })
    return data
